@@ -1,6 +1,7 @@
 /*
  * Copyright © 2017-2018 Red Hat Inc.
  * Copyright © 2018 Jonas Ådahl
+ * Copyright © 2019 Christian Rauch
  *
  * Permission is hereby granted, free of charge, to any person obtaining
  * a copy of this software and associated documentation files (the
@@ -41,6 +42,12 @@
 #include "xdg-shell-client-protocol.h"
 #include "xdg-decoration-client-protocol.h"
 
+#ifdef HAVE_XDG_SHELL_V6
+#define XDG_WM_BASE_VER 6
+#else
+#define XDG_WM_BASE_VER 2
+#endif
+
 struct libdecor {
 	int ref_count;
 
@@ -116,10 +123,6 @@ struct libdecor_frame_private {
 
 	enum libdecor_window_state window_state;
 
-	/* stored dimensions of the floating state */
-	int floating_width;
-	int floating_height;
-
 	enum zxdg_toplevel_decoration_v1_mode decoration_mode;
 
 	enum libdecor_capabilities capabilities;
@@ -249,36 +252,37 @@ frame_get_window_size_for(struct libdecor_frame *frame,
 	struct libdecor *context = frame_priv->context;
 	struct libdecor_plugin *plugin = context->plugin;
 
-	if (frame_has_visible_client_side_decoration(frame)) {
-		return plugin->priv->iface->frame_get_window_size_for(
-					plugin, frame, state,
-					window_width, window_height);
-	} else {
-		*window_width = state->content_width;
-		*window_height = state->content_height;
-		return true;
+	*window_width = state->content_width;
+	*window_height = state->content_height;
+
+	if (frame_has_visible_client_side_decoration(frame) &&
+	    plugin->priv->iface->frame_get_border_size) {
+		int left, right, top, bottom;
+		if (!plugin->priv->iface->frame_get_border_size(
+			plugin, frame, NULL, &left, &right, &top, &bottom))
+			return false;
+		*window_width += left + right;
+		*window_height += top + bottom;
 	}
+
+	return true;
 }
 
-static bool
-window_size_to_content_size(struct libdecor_configuration *configuration,
-			    struct libdecor_frame *frame,
-			    int *content_width,
-			    int *content_height)
+static void
+frame_set_window_geometry(struct libdecor_frame *frame,
+			  int32_t content_width, int32_t content_height)
 {
-	struct libdecor_frame_private *frame_priv = frame->priv;
-	struct libdecor *context = frame_priv->context;
-	struct libdecor_plugin *plugin = context->plugin;
+	struct libdecor_plugin *plugin = frame->priv->context->plugin;
+	int x, y, width, height;
+	int left, right, top, bottom;
 
-	if (frame_has_visible_client_side_decoration(frame)) {
-		return plugin->priv->iface->configuration_get_content_size(
-					plugin, configuration, frame,
-					content_width, content_height);
-	} else {
-		*content_width = configuration->window_width;
-		*content_height = configuration->window_height;
-		return true;
-	}
+	plugin->priv->iface->frame_get_border_size(plugin, frame, NULL,
+						   &left, &right, &top, &bottom);
+	x = -left;
+	y = -top;
+	width = content_width + left + right;
+	height = content_height + top + bottom;
+	xdg_surface_set_window_geometry(frame->priv->xdg_surface, x, y, width, height);
 }
 
 LIBDECOR_EXPORT bool
@@ -287,36 +291,9 @@ libdecor_configuration_get_content_size(struct libdecor_configuration *configura
 					int *width,
 					int *height)
 {
-	int content_width;
-	int content_height;
+	struct libdecor_plugin *plugin = frame->priv->context->plugin;
 
-	if (!configuration->has_size)
-		return false;
-
-	if (configuration->window_width == 0 || configuration->window_height == 0)
-		return false;
-
-	if (!window_size_to_content_size(configuration,
-					 frame,
-					 &content_width,
-					 &content_height))
-		return false;
-
-	*width = content_width;
-	*height = content_height;
-
-	if (state_is_floating(configuration->window_state)) {
-		constrain_content_size(frame, width, height);
-	}
-
-	return true;
-}
-
-LIBDECOR_EXPORT bool
-libdecor_configuration_get_window_size(struct libdecor_configuration *configuration,
-				       int *width,
-				       int *height)
-{
+	/* get configured toplevel dimensions */
 	if (!configuration->has_size)
 		return false;
 
@@ -325,6 +302,27 @@ libdecor_configuration_get_window_size(struct libdecor_configuration *configurat
 
 	*width = configuration->window_width;
 	*height = configuration->window_height;
+
+	/* remove plugin-specific border size */
+	if (frame_has_visible_client_side_decoration(frame) &&
+	    plugin->priv->iface->frame_get_border_size) {
+		int left, right, top, bottom;
+
+		/* Update window state for correct border size calculation */
+		frame->priv->window_state = configuration->window_state;
+		if (!plugin->priv->iface->frame_get_border_size(
+			plugin, frame, configuration, &left, &right, &top, &bottom))
+			return false;
+
+		*width -= (left + right);
+		*height -= (top + bottom);
+	}
+
+	/* constrain content dimensions manually */
+	if (state_is_floating(configuration->window_state)) {
+		constrain_content_size(frame, width, height);
+	}
+
 	return true;
 }
 
@@ -398,6 +396,11 @@ parse_states(struct wl_array *states)
 		case XDG_TOPLEVEL_STATE_TILED_BOTTOM:
 			pending_state |= LIBDECOR_WINDOW_STATE_TILED_BOTTOM;
 			break;
+#ifdef HAVE_XDG_SHELL_V6
+		case XDG_TOPLEVEL_STATE_SUSPENDED:
+			pending_state |= LIBDECOR_WINDOW_STATE_SUSPENDED;
+			break;
+#endif
 		default:
 			break;
 		}
@@ -422,27 +425,8 @@ xdg_toplevel_configure(void *user_data,
 	frame_priv->pending_configuration = libdecor_configuration_new();
 
 	frame_priv->pending_configuration->has_size = true;
-	if (width == 0 && height == 0) {
-		/* client needs to determine window dimensions
-		 * This might happen at the first configuration or after an
-		 * unmaximizing request. In any case, we will forward the stored
-		 * unmaximized state, which will either contain values stored
-		 * by the maximizing request, or 0.
-		 */
-		frame_priv->pending_configuration->window_width =
-				frame->priv->floating_width;
-		frame_priv->pending_configuration->window_height =
-				frame->priv->floating_height;
-	} else {
-		/* store current floating state */
-		if (state_is_floating(window_state)) {
-			frame->priv->floating_width = width;
-			frame->priv->floating_height = height;
-		}
-
-		frame_priv->pending_configuration->window_width = width;
-		frame_priv->pending_configuration->window_height = height;
-	}
+	frame_priv->pending_configuration->window_width = width;
+	frame_priv->pending_configuration->window_height = height;
 
 	frame_priv->pending_configuration->has_window_state = true;
 	frame_priv->pending_configuration->window_state = window_state;
@@ -458,9 +442,30 @@ xdg_toplevel_close(void *user_data,
 	frame_priv->iface->close(frame, frame_priv->user_data);
 }
 
+#ifdef HAVE_XDG_SHELL_V6
+static void
+xdg_toplevel_configure_bounds(void *data,
+			      struct xdg_toplevel *xdg_toplevel,
+			      int32_t width,
+			      int32_t height)
+{
+}
+
+static void
+xdg_toplevel_wm_capabilities(void *data,
+			     struct xdg_toplevel *xdg_toplevel,
+			     struct wl_array *capabilities)
+{
+}
+#endif
+
 static const struct xdg_toplevel_listener xdg_toplevel_listener = {
 	xdg_toplevel_configure,
 	xdg_toplevel_close,
+#ifdef HAVE_XDG_SHELL_V6
+	xdg_toplevel_configure_bounds,
+	xdg_toplevel_wm_capabilities,
+#endif
 };
 
 static void
@@ -600,6 +605,13 @@ libdecor_frame_unref(struct libdecor_frame *frame)
 		struct libdecor *context = frame_priv->context;
 		struct libdecor_plugin *plugin = context->plugin;
 
+		if (context->decoration_manager && frame_priv->toplevel_decoration) {
+        		zxdg_toplevel_decoration_v1_destroy(frame_priv->toplevel_decoration);
+        		frame_priv->toplevel_decoration = NULL;
+        	}
+
+		wl_list_remove(&frame->link);
+
 		if (frame_priv->xdg_toplevel)
 			xdg_toplevel_destroy(frame_priv->xdg_toplevel);
 		if (frame_priv->xdg_surface)
@@ -653,12 +665,12 @@ libdecor_frame_set_visibility(struct libdecor_frame *frame,
 	} else {
 		/* destroy client-side decorations */
 		plugin->priv->iface->frame_free(plugin, frame);
-
-		libdecor_frame_set_window_geometry(frame, 0, 0,
-						   frame_priv->content_width,
-						   frame_priv->content_height);
 	}
 
+	frame_set_window_geometry(frame,
+				  frame_priv->content_width,
+				  frame_priv->content_height);
+
 	libdecor_frame_toplevel_commit(frame);
 }
 
@@ -698,7 +710,7 @@ libdecor_frame_set_title(struct libdecor_frame *frame,
 			return;
 
 		xdg_toplevel_set_title(frame_priv->xdg_toplevel, title);
-	
+
 		plugin->priv->iface->frame_property_changed(plugin, frame);
 	}
 }
@@ -854,20 +866,17 @@ libdecor_frame_translate_coordinate(struct libdecor_frame *frame,
 	struct libdecor *context = frame_priv->context;
 	struct libdecor_plugin *plugin = context->plugin;
 
-	plugin->priv->iface->frame_translate_coordinate(plugin, frame,
-							content_x, content_y,
-							frame_x, frame_y);
-}
+	*frame_x = content_x;
+	*frame_y = content_y;
 
-LIBDECOR_EXPORT void
-libdecor_frame_set_max_content_size(struct libdecor_frame *frame,
-				    int content_width,
-				    int content_height)
-{
-	struct libdecor_frame_private *frame_priv = frame->priv;
-
-	frame_priv->state.content_limits.max_width = content_width;
-	frame_priv->state.content_limits.max_height = content_height;
+	if (frame_has_visible_client_side_decoration(frame) &&
+	    plugin->priv->iface->frame_get_border_size) {
+		int left, top;
+		plugin->priv->iface->frame_get_border_size(plugin, frame, NULL,
+							   &left, NULL, &top, NULL);
+		*frame_x += left;
+		*frame_y += top;
+	}
 }
 
 LIBDECOR_EXPORT void
@@ -882,11 +891,36 @@ libdecor_frame_set_min_content_size(struct libdecor_frame *frame,
 }
 
 LIBDECOR_EXPORT void
-libdecor_frame_set_window_geometry(struct libdecor_frame *frame,
-				   int32_t x, int32_t y,
-				   int32_t width, int32_t height)
+libdecor_frame_set_max_content_size(struct libdecor_frame *frame,
+				    int content_width,
+				    int content_height)
 {
-	xdg_surface_set_window_geometry(frame->priv->xdg_surface, x, y, width, height);
+	struct libdecor_frame_private *frame_priv = frame->priv;
+
+	frame_priv->state.content_limits.max_width = content_width;
+	frame_priv->state.content_limits.max_height = content_height;
+}
+
+LIBDECOR_EXPORT void
+libdecor_frame_get_min_content_size(const struct libdecor_frame *frame,
+				    int *content_width,
+				    int *content_height)
+{
+	struct libdecor_frame_private *frame_priv = frame->priv;
+
+	*content_width = frame_priv->state.content_limits.min_width;
+	*content_height = frame_priv->state.content_limits.min_height;
+}
+
+LIBDECOR_EXPORT void
+libdecor_frame_get_max_content_size(const struct libdecor_frame *frame,
+				    int *content_width,
+				    int *content_height)
+{
+	struct libdecor_frame_private *frame_priv = frame->priv;
+
+	*content_width = frame_priv->state.content_limits.max_width;
+	*content_height = frame_priv->state.content_limits.max_height;
 }
 
 LIBDECOR_EXPORT enum libdecor_capabilities
@@ -1121,18 +1155,11 @@ libdecor_frame_commit(struct libdecor_frame *frame,
 						  configuration);
 	} else {
 		plugin->priv->iface->frame_free(plugin, frame);
-
-		libdecor_frame_set_window_geometry(frame, 0, 0,
-						   frame_priv->content_width,
-						   frame_priv->content_height);
 	}
 
-	/* set the floating dimensions via the application's requested content size */
-	if (configuration == NULL) {
-		frame_get_window_size_for(frame, state,
-					  &frame->priv->floating_width,
-					  &frame->priv->floating_height);
-	}
+	frame_set_window_geometry(frame,
+				  frame_priv->content_width,
+				  frame_priv->content_height);
 
 	if (configuration) {
 		xdg_surface_ack_configure(frame_priv->xdg_surface,
@@ -1225,6 +1252,7 @@ libdecor_plugin_init(struct libdecor_plugin *plugin,
 LIBDECOR_EXPORT void
 libdecor_plugin_release(struct libdecor_plugin *plugin)
 {
+	free(plugin->priv);
 }
 
 static void
@@ -1247,7 +1275,7 @@ init_xdg_wm_base(struct libdecor *context,
 	context->xdg_wm_base = wl_registry_bind(context->wl_registry,
 						id,
 						&xdg_wm_base_interface,
-						MIN(version,2));
+						MIN(version,XDG_WM_BASE_VER));
 	xdg_wm_base_add_listener(context->xdg_wm_base,
 				 &xdg_wm_base_listener,
 				 context);
@@ -1265,6 +1293,12 @@ registry_handle_global(void *user_data,
 	if (!strcmp(interface, xdg_wm_base_interface.name)) {
 		init_xdg_wm_base(context, id, version);
 	} else if (!strcmp(interface, zxdg_decoration_manager_v1_interface.name)) {
+		const char *force_csd = getenv("LIBDECOR_FORCE_CSD");
+
+		if (force_csd && atoi(force_csd)) {
+			return;
+		}
+
 		context->decoration_manager = wl_registry_bind(
 				context->wl_registry, id,
 				&zxdg_decoration_manager_v1_interface,
@@ -1391,18 +1425,44 @@ calculate_priority(const struct libdecor_plugin_description *plugin_description)
 	return -1;
 }
 
+static bool
+check_symbol_conflicts(const struct libdecor_plugin_description *plugin_description, void *lib)
+{
+	char * const *symbol;
+
+	symbol = plugin_description->conflicting_symbols;
+	while (*symbol) {
+		dlerror();
+		void *sym = dlsym(RTLD_DEFAULT, *symbol);
+		if (!dlerror()) {
+			void *libsym = dlsym(lib, *symbol);
+			if (!dlerror() && libsym != sym) {
+				fprintf(stderr, "Plugin \"%s\" uses conflicting symbol \"%s\".\n",
+					plugin_description->description, *symbol);
+				return false;
+			}
+		}
+
+		symbol++;
+	}
+
+	return true;
+}
+
 static struct plugin_loader *
 load_plugin_loader(struct libdecor *context,
 		   const char *path,
 		   const char *name)
 {
+	char *ext;
 	char *filename;
 	void *lib;
 	const struct libdecor_plugin_description *plugin_description;
 	int priority;
 	struct plugin_loader *plugin_loader;
 
-	if (!strstr(name, ".so"))
+	ext = strrchr(name, '.');
+	if (ext == NULL || strcmp(ext, ".so") != 0)
 		return NULL;
 
 	if (asprintf(&filename, "%s/%s", path, name) == -1)
@@ -1417,21 +1477,21 @@ load_plugin_loader(struct libdecor *context,
 
 	plugin_description = dlsym(lib, "libdecor_plugin_description");
 	if (!plugin_description) {
-		dlclose(lib);
 		fprintf(stderr,
 			"Failed to load plugin '%s': no plugin description symbol\n",
 			name);
+		dlclose(lib);
 		return NULL;
 	}
 
 	if (plugin_description->api_version != LIBDECOR_PLUGIN_API_VERSION) {
-		dlclose(lib);
 		fprintf(stderr,
 			"Plugin '%s' found, but it's incompatible "
 			"(expected API version %d, but got %d)\n",
 			name,
 			LIBDECOR_PLUGIN_API_VERSION,
 			plugin_description->api_version);
+		dlclose(lib);
 		return NULL;
 	}
 
@@ -1440,12 +1500,17 @@ load_plugin_loader(struct libdecor *context,
 		return NULL;
 	}
 
+	if (!check_symbol_conflicts(plugin_description, lib)) {
+		dlclose(lib);
+		return NULL;
+	}
+
 	priority = calculate_priority(plugin_description);
 	if (priority == -1) {
-		dlclose(lib);
 		fprintf(stderr,
 			"Plugin '%s' found, but has an invalid description\n",
 			name);
+		dlclose(lib);
 		return NULL;
 	}
 
@@ -1468,42 +1533,52 @@ plugin_loader_higher_priority(struct plugin_loader *plugin_loader,
 static int
 init_plugins(struct libdecor *context)
 {
-	const char *plugin_dir;
+	const char *plugin_dir_env;
+	char *all_plugin_dirs;
+	char *plugin_dir;
+	char *saveptr;
 	DIR *dir;
 	struct wl_list plugin_loaders;
 	struct plugin_loader *plugin_loader, *tmp;
 	struct plugin_loader *best_plugin_loader;
 	struct libdecor_plugin *plugin;
 
-	plugin_dir = getenv("LIBDECOR_PLUGIN_DIR");
-	if (!plugin_dir)
-		plugin_dir = LIBDECOR_PLUGIN_DIR;
-
-	dir = opendir(plugin_dir);
-	if (!dir) {
-		fprintf(stderr, "Couldn't open plugin directory: %s\n",
-			strerror(errno));
-		return -1;
+	plugin_dir_env = getenv("LIBDECOR_PLUGIN_DIR");
+	if (!plugin_dir_env) {
+		plugin_dir_env = LIBDECOR_PLUGIN_DIR;
 	}
+	all_plugin_dirs = strdup(plugin_dir_env);
 
 	wl_list_init(&plugin_loaders);
+	plugin_dir = strtok_r(all_plugin_dirs, ":", &saveptr);
+	while (plugin_dir) {
+		dir = opendir(plugin_dir);
+		if (!dir) {
+			fprintf(stderr, "Couldn't open plugin directory: %s\n",
+				strerror(errno));
+		} else {
 
-	while (true) {
-		struct dirent *de;
-		struct plugin_loader *plugin_loader;
+			while (true) {
+				struct dirent *de;
 
-		de = readdir(dir);
-		if (!de)
-			break;
+				de = readdir(dir);
+				if (!de)
+					break;
 
-		plugin_loader = load_plugin_loader(context, plugin_dir, de->d_name);
-		if (!plugin_loader)
-			continue;
+				plugin_loader = load_plugin_loader(context,
+								   plugin_dir,
+								   de->d_name);
+				if (!plugin_loader)
+					continue;
 
-		wl_list_insert(plugin_loaders.prev, &plugin_loader->link);
+				wl_list_insert(plugin_loaders.prev, &plugin_loader->link);
+			}
+
+			closedir(dir);
+		}
+		plugin_dir = strtok_r(NULL, ":", &saveptr);
 	}
-
-	closedir(dir);
+	free(all_plugin_dirs);
 
 retry_next:
 	best_plugin_loader = NULL;
